import secrets
from ..models.accessToken import AccessToken
from flaschengeist.system.database import db
from flaschengeist import logger
from werkzeug.exceptions import Forbidden
from datetime import datetime, timezone
from . import Singleton


class AccessTokenController(metaclass=Singleton):
    """ Control all created AccessToken

        This Class create, delete, find and manage AccessToken.

        Attributes:
            lifetime: Variable for the Lifetime of one AccessToken in seconds.
    """

    def __init__(self, lifetime=1800):
        self.lifetime = lifetime

    def validate_token(self, token, user_agent, permissions):
        """ Verify access token

            Verify an AccessToken and Roles so if the User has permission or not.
            Retrieves the access token if valid else retrieves False

            Args:
                token: Token to verify.
                user_agent: User agent of browser to check
                permissions: Permissions needed to access restricted routes
            Returns:
                An the AccessToken for this given Token or False.
        """
        logger.debug("check token {{ {} }} is valid".format(token))
        access_token = AccessToken.query.filter_by(token=token).one_or_none()
        if access_token:
            logger.debug("token found, check if expired or invalid user agent differs")
            if access_token.expires >= datetime.utcnow() and (
                    access_token.browser == user_agent.browser and
                    access_token.platform == user_agent.platform):
                if not permissions or access_token.user.has_permissions(permissions):
                    access_token.refresh()
                    db.session.commit()
                    return access_token
            else:
                logger.debug("access token is out of date or invalid client used")
                self.delete_token(access_token)
        logger.debug("no valid access token with token: {{ {} }} and permissions: {{ {} }}".format(token, permissions))
        return False

    def create(self, user, user_agent=None) -> AccessToken:
        """ Create an AccessToken

        Args:
            user: For which User is to create an AccessToken
            user_agent: User agent to identify session

        Returns:
                AccessToken: A created Token for User
        """
        logger.debug("create access token")
        token_str = secrets.token_hex(16)
        token = AccessToken(token=token_str, user=user, lifetime=self.lifetime,
                            browser=user_agent.browser, platform=user_agent.platform)
        token.refresh()
        db.session.add(token)
        db.session.commit()

        logger.debug("access token is {{ {} }}".format(token))
        return token

    def get_token(self, token, owner=None):
        """Retrieves AccessToken from token string

        Args:
            token (str): Token string
            owner (User, optional): User owning the token

        Raises:
            Forbidden: Raised if owner is set but does not match
        Returns:
            AccessToken: Token object identified by given token string
        """
        access_token = AccessToken.query.filter(AccessToken.token == token).one_or_none()
        if access_token and (owner and owner != access_token.user):
            raise Forbidden
        return access_token

    def get_users_tokens(self, user):
        return AccessToken.query.filter(AccessToken.user == user)

    @staticmethod
    def delete_token(token: AccessToken):
        """Deletes given AccessToken

        Args:
            token (AccessToken): Token to delete
        """
        db.session.delete(token)
        db.session.commit()

    @staticmethod
    def update_token(token):
        token.refresh()
        db.session.commit()

    def set_lifetime(self, token, lifetime):
        token.lifetime = lifetime
        self.update_token(token)

    def clear_expired(self):
        """Remove expired tokens from database"""
        logger.debug("Clear expired AccessToken")
        deleted = AccessToken.query.filter(AccessToken.expires < datetime.utcnow()).delete()
        logger.debug("{} tokens have been removed".format(deleted))
        db.session.commit()
